/*
 * DrawPanel.java
 *
 * Created on 8. jen 2005, 18:49
 */

package gui;
import gui.parts.SizeChange;
import gui.parts.SizeChangeListener;
import java.awt.Color;
import java.io.IOException;
import javax.imageio.ImageIO;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.io.File;
import java.util.ArrayList;
import java.util.List;
import org.jibble.epsgraphics.EpsGraphics2D;
import gui.parts.Part;

/**
 * Trida, ktera vykresluje graficke komponenty, potomky tridy {@link gui.parts.Part }
 * @author Ondrej Hanousek
 */
public class DrawPanel extends javax.swing.JComponent{
    
    private AffineTransform tx = new AffineTransform();
    private int height=0;
    private int width=0;
    private double zoom=1.0;
    private List<Part> parts=new ArrayList<Part>();
    
    
    /**
     * Vytvori panel.
     */
    public DrawPanel() {
    }
    
    
    /**
     * Vraci velikost vnitrnich komponent.
     * @return Minimalni obdelnik opisujici graficke objekty
     */
    public Dimension getPreferredSize(){return new Dimension(width,height);}
    
    /**
     * Velikost obsazenych grafickych objektu.
     * @return Minimalni obdelnik opisujici graficke objekty
     */
    public Dimension getMinimumSize(){return new Dimension(width,height);}
    
    /**
     * Prida objekt tridy {@link gui.parts.Part }, ktery ma byt vykreslovan
     * @param p objekt, ktery ma byt vykreslovan
     */
    public void addPart(Part p){
        parts.add(p);
        p.addSizeChangeEventListener(new SizeChangeListener(){
            public void sizeChanged(SizeChange evt){
                recomputeSize();
            }
        });
        recomputeSize();
        revalidate();
    }
    
    /**
     * Prida objekt tridy {@link gui.parts.Part }, ktery nema byt vykreslovan
     * @param p objekt, ktery ma byt vykreslovan
     */
    public void removePart(Part p){
        if (p==null) return ;
        Rectangle bounds=p.getBounds();
        parts.remove(p);
        recomputeSize();
        revalidate();
    }
    
    private void recomputeSize(){
      //  System.out.println("Recomputing");
        width=0;
        height=0;
        int x=Integer.MAX_VALUE;
        int y=Integer.MAX_VALUE;
        for (Part part:parts) {
            
            Rectangle tmp=part.getBounds();
            width=Math.max(width,tmp.x+tmp.width);
            height=Math.max(height,tmp.y+tmp.height);
            x=Math.min(x,tmp.x);
            y=Math.min(y,tmp.y);
         //   System.out.println(x+" "+y+" "+width+" "+height);
        }
       //  if (x<0 || y<0) {
            for (Part part:parts) {
                part.move(x,y);
           }
            width=width-x;
            height=height-y;
        // }
       // System.out.println(width+ " "+ height);
        super.setSize(width+x,height+y);
    }
    
    /**
     * Nastavi novou kolekci objektu, ktere se maji vykreslovat.
     * @param parts kolekce vykreslovanych objektu
     */
    public void setList( List<Part> parts){
        this.parts=parts;
        recomputeSize();
        
    }
    
    /**
     * Vymaze obsah panelu, vprazdni kolekci zobrazovanych objektu.
     */
    public void reset(){
        parts=new ArrayList<Part>();
        width=10;
        height=10;
        revalidate();
    }
    
    
    /**
     * Paints speciffied rectangle only - ddrawn objects are cliped
     * by speciffied rectangle.
     * @param g Graphics content
     * @param area clip area
     */
    public void paintArea(Graphics g, Rectangle area) {
        
        Graphics2D g2d = (Graphics2D) g;
        g2d.scale(zoom,zoom);
        Shape view=new Rectangle2D.Double(area.x,area.y,area.width,area.height);
        for (Part p:parts) {
            p.paintPart(g2d);
       //   g2d.draw(p.getBounds());
        }
       
       g2d.setClip(area);
    //    g2d.drawRect(2,2,width-4,height-4);
        
    }
    
    /**
     * method called when repaint necessary
     * @param g Graphical content
     */
    public void paint(Graphics g) {
        Rectangle r=getVisibleRect();
        paintArea(g, r);
       // System.out.println(r);
    }
    
    /**
     * Ulozi do soboru rastrovy obrazek, ktery odpovida zobrazenee grafice.
     * @param outFile souber pro vystup
     * @return 0 - vystup ok
     * -1 - chyba
     */
    public int writePngImage(File outFile){
        recomputeSize();
        Rectangle all=new Rectangle(0,0, width,height);
        BufferedImage image=new BufferedImage(width,height,BufferedImage.TYPE_3BYTE_BGR);
        Graphics g=image.getGraphics();
        g.setColor(Color.WHITE);
        g.fillRect(0,0, width, height);
        paintArea(g, all);
        try {
            ImageIO.write(image, "png", outFile);
        } catch (IOException ex) {
            System.err.println(ex.getLocalizedMessage());
            return -1;
        } finally {
            g.dispose();
        }
        return 0;
        
    }
    
    /**
     * Vypise vystup ve vektorovem formatu PS.
     * @return 0 - vystup ok
     * -1 - chyba vystupu
     * @param colorMode barevny mod - barevne, cernobile, stupne sede
     * @param outFile soubor pro ulozeni dat
     */
    public int writePsImage(File outFile, int colorMode){
        recomputeSize();
        
        try {
            Rectangle all=new Rectangle(0,0, width,height);
           EpsGraphics2D out=new EpsGraphics2D("Komprese",outFile, 0, 0, width, height);
            out.setColorDepth(colorMode);
            paintArea(out, all);
            out.flush();
            out.close();
        } catch (IOException ex) {
            System.err.println(ex.getLocalizedMessage());
            return -1;
        }
        
        return 0;
    }
    
    
}

