/*
 * SFOutputStream.java
 * Created on 3. z 2005, 12:06
 */

package streams;
import hanus.FibonacciEncoder;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FilterOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Arrays;
import pradny.BitOutputStream;
import pradny.UnsupportedOutputException;
import streams.utils.ByteInfo;
import streams.utils.ByteInfoCountComparator;
import streams.utils.ByteInfoDataComparator;


/**
 * Trida, ktera umoznuje zakodovani dat Shanon-Fanovym kodovanim.
 * Pouziva se semiadaptivni verze algoritmu, datovy model je
 * odeslan jako cetnosti jednotlivych bytu zakodovanych Fibonacciho kodem
 * radu 2.
 * @author Ondrej Hanousek
 */
public class SFOutputStream extends FilterOutputStream {
    
    private ByteInfo[] byteTable=new ByteInfo[256];
    private boolean finished=false;
    private BitOutputStream bos;
    private FibonacciEncoder enc;
    private long size=0;
    private File tempFile;
    private BufferedOutputStream tempOutBuffer;
    
    /**
     * Vytvori tridu - stream - ktera bude zapisovat do zadaneho streamu 
     * kodovana data.
     * @param os vystupni proud
     * @throws pradny.UnsupportedOutputException if unsuitable output stream used
     * @throws java.io.IOException if a I/O error has occurred
     */
    public SFOutputStream(OutputStream os) throws UnsupportedOutputException, IOException  {
        super(os);
        bos=new BitOutputStream(os);
        enc=new FibonacciEncoder(bos);
        tempFile = File.createTempFile("SFStream", ".tmp");
        tempFile.deleteOnExit();
        tempOutBuffer=new BufferedOutputStream(new FileOutputStream(tempFile));
        for (int i=0;i<256;i++){
            byteTable[i]=new ByteInfo(i);
            
        }
        
    }
    
    /**
     * Metoda pro zapis dat do streamu - data jsou kopirovana 
     * do temp souboru a jsou pocitany pocty bajtu.
     * @throws java.io.IOException pri i/o chybe
     * @param b data k zapisu (byte)
     */
    public synchronized void write(int b) throws java.io.IOException {
        if (!finished) {
            byteTable[b].increment();
            size++;
            tempOutBuffer.write(b);
            
        } else { 
            throw new IOException("Stream finished!");
            
        }
    }
    
    
    /**
     * Zavolanim teto metody se vytvori kody pro 
     * jednotlive bajty a do vystupniho proudu je zapsan
     * statisticky model (cetnosti bajtu) a zakodovana data.
     * Po provedeni flush() neni dalsi zapis mozny.
     * @throws java.io.IOException pri I/O chybe
     */
    public void flush() throws java.io.IOException {
        finish();
        super.flush();
    }
    
    /**
     * Finishes writing compressed data to the output stream without closing the underlying stream.
     * No more data writting possible.
     * @throws java.io.IOException if an I/O error has occurred
     */
    private void finish() throws IOException {
        if (!finished) {
            finished=true;
            writeCompressed();
        }
    }
    
    private  void writeCompressed() throws IOException {
        tempOutBuffer.flush();
 //       BufferedInputStream tempInput=new BufferedInputStream(new FileInputStream(tempFile));
        
        
        Arrays.sort(byteTable,new ByteInfoCountComparator());
        doCodes(byteTable,0, 256, size);
//       System.out.println(Arrays.toString(byteTable));
        Arrays.sort(byteTable, new ByteInfoDataComparator());
        
        for (ByteInfo info:byteTable) {
            enc.encodeLong(info.getCount());
        }
        
        BufferedInputStream bis=new BufferedInputStream(new FileInputStream(tempFile));
        
        int data;
        while ((data=bis.read())!=-1) {
            bos.write(byteTable[data].getCodeBits());
        }
        bis.close();
        bos.close();
    }
    
    /**
     * Uzavreni proudu. Provede se i metoda flush().
     * @throws java.io.IOException if an I/O error has occcured
     */
    public void close() throws java.io.IOException {
        finish();
        super.close();
        enc.close();
        bos.close();
    }
    
    /**
     * Makes code for each byte in bytetable.
     * @param byteTable Table with information about byte - must be sorted by count
     *
     * @param from begin index in byteTable array, used in recursive calls
     * @param to end index in byteTable array, used in recursive calls
     * @param size count of bytes in from - to byteTable range
     */
protected static void doCodes(ByteInfo[] byteTable, int from, int to, long size){
       //System.out.println(from +" - " +to + " -  "+size);
        if ((to-from)<=1) {
          //  System.out.println("ret");
            return;
        }
        
        for (int i=from;i<to;i++){
            from=i;
              if (to-from<=1) {
                    byteTable[i].addBit(false);
                    return ;
                }
            
            if (byteTable[i].getCount()!=0) {
                break;
            }
           
        }
        
        int splitPoint=from;
        long halfSize=byteTable[from].getCount();
        
        for (int i=from;i<to;i++) {
            
            byteTable[i].addBit(true);
            long lastCount=byteTable[i+1].getCount();
            halfSize+=lastCount;
            if (halfSize>=size/2.0) {
                if ((halfSize-size/2.0)<(lastCount/2.0)) {
                    byteTable[i+1].addBit(true);
                    splitPoint=i+2;
                    break;
                } else {
                    splitPoint=i+1;
                    halfSize-=lastCount;
                    break;
                }
            }
        }
        for (int i=splitPoint;i<to;i++) {
            byteTable[i].addBit(false);
        }
        
        doCodes(byteTable, from, splitPoint, halfSize);
        doCodes(byteTable, splitPoint, to, size-halfSize);
        
    }
    
}
