/*
 * SHOutputStream.java
 *
 * Created on 25. jen 2005, 22:52
 */

package streams;

import hanus.FibonacciEncoder;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FilterOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Arrays;
import java.util.PriorityQueue;
import pradny.BitOutputStream;
import pradny.UnsupportedOutputException;
import streams.utils.ByteInfo;
import streams.utils.ByteInfoCountComparator;
import streams.utils.ByteInfoDataComparator;
import streams.utils.ByteInfoSH;

/**
 * Trida, ktera umoznuje zakodovani dat statickym Huffmanovym kodovanim.
 * Pouziva se semiadaptivni verze algoritmu, datovy model je
 * odeslan jako cetnosti jednotlivych bytu zakodovanych Fibonacciho kodem
 * radu 2.
 * @author Ondrej Hanousek
 */
public class SHOutputStream extends FilterOutputStream {
    
    private ByteInfoSH[] byteTable=new ByteInfoSH[256];
    private boolean finished=false;
    private BitOutputStream bos;
    private FibonacciEncoder enc;
    private long size=0;
    private File tempFile;
    private BufferedOutputStream tempOutBuffer;
    /**
     * Vytvori koder statickeho Huffmanova kodu.
     * @param os vystup pro data
     * @throws pradny.UnsupportedOutputException pri nevhodnem vystupu
     * @throws java.io.IOException pri I/O chybe
     */
    public SHOutputStream(OutputStream os) throws UnsupportedOutputException, IOException  {
        super(os);
        bos=new BitOutputStream(os);
        enc=new FibonacciEncoder(bos);
        tempFile = File.createTempFile("SHStream", ".tmp");
        tempFile.deleteOnExit();
        tempOutBuffer=new BufferedOutputStream(new FileOutputStream(tempFile));
        for (int i=0;i<256;i++){
            byteTable[i]=new ByteInfoSH(i);
        }
    }
    
    /**
     * Zapise data - vytvari statisticky model a zapisovana data uklada
     * do docasneho souboru.
     * @param b zapisovana data
     * @throws java.io.IOException pri I/O chybe
     */
    public synchronized void write(int b) throws java.io.IOException {
        if (!finished) {
            byteTable[b].increment();
            size++;
            tempOutBuffer.write(b);
        } else
        {
            throw new IOException("Stream finished!");
        }
    }
    
    
/**
 * Vypise do vystupu statisticky model, vytvori kody znaku a zapise zakodovana data.
 * Po flush uz neni dalsi zapis mozny.
 * @throws java.io.IOException pri I/O chybe
 */
public void flush()  throws IOException {
    finish();
    super.flush();
}

private void finish() throws IOException {
        if (!finished) {
            finished=true;
            writeCompressed();
        }
    }
    
    
    private  void writeCompressed() throws IOException {
        tempOutBuffer.flush();
 //       BufferedInputStream tempInput=new BufferedInputStream(new FileInputStream(tempFile));
        
        
        Arrays.sort(byteTable,new ByteInfoCountComparator());

        buildHuffmanTree(byteTable);

        Arrays.sort(byteTable, new ByteInfoDataComparator());
        
        for (ByteInfo info:byteTable) {
            enc.encodeLong(info.getCount());
        }
        
        BufferedInputStream bis=new BufferedInputStream(new FileInputStream(tempFile));
        
        int data;
        while ((data=bis.read())!=-1) {
            bos.write(byteTable[data].getCodeBits());
       //     System.out.println(Arrays.toString(byteTable[data].getCodeBits()));
        }

        bis.close();
        bos.close();
    }
    
   protected static ByteInfoSH buildHuffmanTree(ByteInfoSH[] byteTable){
        int k=0;
        PriorityQueue<ByteInfoSH> queue=new PriorityQueue<ByteInfoSH>(256, new ByteInfoCountComparator());
        for (int i=0;i<byteTable.length;i++){
            if (byteTable[i].getCount()!=0) {
                k++;
                byteTable[i].setLeaf(true);
                queue.add(byteTable[i]);
            }
        }
        while (k>=2){
            ByteInfoSH first=queue.poll();
            ByteInfoSH second=queue.poll();
            queue.add(new ByteInfoSH(first,second));
            k--;
        }
        return queue.peek();
    }

    /**
     * Provede flush a uzavre stream.
     * @throws java.io.IOException pri I/O chybe
     */
    public void close() throws IOException {
        finish();
        super.close();
        enc.close();
        bos.close();
    }
}

